import { isSameToken, } from "@wormhole-foundation/sdk-definitions";
import { TokenTransfer } from "../protocols/tokenTransfer";
import { TransferState, isSourceFinalized } from "../wormholeTransfer";
import { ManualRoute } from "./route";
export class TokenBridgeRoute extends ManualRoute {
    async isSupported() {
        // No transfers to same chain
        if (this.request.from.chain === this.request.to.chain)
            return false;
        // No transfers to unsupported chains
        const fromChain = this.wh.getChain(this.request.from.chain);
        if (!fromChain.supportsTokenBridge())
            return false;
        const toChain = this.wh.getChain(this.request.to.chain);
        if (!toChain.supportsTokenBridge())
            return false;
        // If the destination token was set, and its different than what
        // we'd get from a token bridge transfer, then this route is not supported
        if (this.request.destination) {
            const destToken = await TokenTransfer.lookupDestinationToken(fromChain, toChain, this.request.source);
            if (!isSameToken(destToken, this.request.destination))
                return false;
        }
        return true;
    }
    async validate(options) {
        const transfer = this.toTransferDetails(options);
        try {
            await TokenTransfer.validateTransferDetails(this.wh, transfer);
            return { valid: true };
        }
        catch (e) {
            return { valid: false, error: e };
        }
    }
    getDefaultOptions() {
        return { payload: undefined };
    }
    async initiate(signer, options) {
        const fromChain = this.wh.getChain(this.request.from.chain);
        const transfer = this.toTransferDetails(options);
        const txids = await TokenTransfer.transfer(fromChain, transfer, signer);
        const msg = await TokenTransfer.getTransferMessage(fromChain, txids[txids.length - 1].txid);
        return {
            protocol: "TokenBridge",
            from: transfer.from.chain,
            to: transfer.to.chain,
            state: TransferState.SourceFinalized,
            request: transfer,
            originTxs: txids,
            attestation: { id: msg },
        };
    }
    async complete(signer, receipt) {
        if (!isSourceFinalized(receipt))
            throw new Error("The source must be finalized in order to complete the transfer");
        const toChain = this.wh.getChain(this.request.to.chain);
        const vaa = await TokenTransfer.getTransferVaa(this.wh, receipt.attestation.id);
        return await TokenTransfer.redeem(toChain, vaa, signer);
    }
    async *track(receipt) {
        return TokenTransfer.track(this.wh, receipt);
    }
    toTransferDetails(options) {
        return {
            token: this.request.source,
            amount: this.request.amount,
            from: this.request.from,
            to: this.request.to,
            ...options,
        };
    }
}
//# sourceMappingURL=tokenBridgeRoute.js.map